import logging
import threading
import os

import jellyfin_apiclient_python.exceptions

from .conf import settings
from .player import playerManager
from .utils import Timer

log = logging.getLogger("timeline")


class TimelineManager(threading.Thread):
    def __init__(self):
        self.idleTimer = Timer()
        self.halt = False
        self.trigger = threading.Event()
        self.is_idle = True

        threading.Thread.__init__(self)

    def stop(self):
        self.halt = True
        self.join()

    def run(self):
        while not self.halt:
            if playerManager.is_active() and (
                not settings.idle_when_paused or not playerManager.is_paused()
            ):
                self.send_timeline()
                self.delay_idle()
            if self.idleTimer.elapsed() > settings.idle_cmd_delay and not self.is_idle:
                if (
                    settings.idle_when_paused
                    and settings.stop_idle
                    and playerManager.has_video()
                ):
                    playerManager.stop()
                if settings.idle_cmd:
                    os.system(settings.idle_cmd)
                self.is_idle = True
            if self.trigger.wait(5):
                self.trigger.clear()

    def delay_idle(self):
        self.idleTimer.restart()
        self.is_idle = False

    @staticmethod
    def send_timeline():
        try:
            # Send_timeline sometimes (once every couple hours) gets a 404 response from Jellyfin.
            # Without this try/except that would cause this entire thread to crash keeping it from self-healing.
            playerManager.send_timeline()
        except jellyfin_apiclient_python.exceptions.HTTPException:
            # FIXME: Log this
            pass


timelineManager = TimelineManager()



from pypresence import Presence
import time

client_id = "743296148592263240"
RPC = Presence(client_id)
RPC.connect()


def send_presence(
    title: str,
    subtitle: str,
    playback_time: float = None,
    duration: float = None,
    playing: bool = False,
):
    small_image = "play-dark3" if playing else None
    start = None
    end = None
    if playback_time is not None and duration is not None and playing:
        start = int(time.time() - playback_time)
        end = int(start + duration)
    RPC.update(
        state=subtitle,
        details=title,
        instance=False,
        large_image="jellyfin2",
        start=start,
        end=end,
        large_text="Jellyfin",
        small_image=small_image,
    )


def clear_presence():
    RPC.clear()



APP_NAME = "jellyfin-mpv-shim"
USER_APP_NAME = "Jellyfin MPV Shim"
CLIENT_VERSION = "1.10.0"
USER_AGENT = "Jellyfin-MPV-Shim/%s" % CLIENT_VERSION
CAPABILITIES = {
        "PlayableMediaTypes": "Video",
        "SupportsMediaControl": True,
        "SupportedCommands": (
            "MoveUp,MoveDown,MoveLeft,MoveRight,Select,"
            "Back,ToggleFullscreen,"
            "GoHome,GoToSettings,TakeScreenshot,"
            "VolumeUp,VolumeDown,ToggleMute,"
            "SetAudioStreamIndex,SetSubtitleStreamIndex,"
            "Mute,Unmute,SetVolume,DisplayContent,"
            "Play,Playstate,PlayNext,PlayMediaSource"
        ),
    }


import requests
import datetime
import logging
import webbrowser

from .constants import CLIENT_VERSION
from .conf import settings
from .i18n import _

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .player import PlayerManager as PlayerManager_type

log = logging.getLogger("update_check")

release_url = "https://github.com/jellyfin/jellyfin-desktop/releases/"
release_urls = [release_url]
one_day = 86400


class UpdateChecker:
    def __init__(self, player_manager: "PlayerManager_type"):
        self.playerManager = player_manager
        self.has_notified = False
        self.new_version = None
        self.last_check = None

    def _check_updates(self):
        log.info("Checking for updates...")
        for release_url in release_urls:
            try:
                response = requests.get(
                    release_url + "latest", allow_redirects=False, timeout=(3, 10)
                )
                if response.status_code != 302:
                    log.warning("Release page returned bad status code.")
                    continue
                if not response.headers["location"].startswith(release_url):
                    log.warning("Release page does not start with release_url.")
                    continue
                version = response.headers["location"][len(release_url) + 5 :]
                if CLIENT_VERSION != version:
                    self.new_version = version
                    break
            except Exception:
                log.error("Could not check for updates.", exc_info=True)
            return self.new_version is not None

    def check(self):
        if not settings.check_updates:
            return

        if (
            self.last_check is not None
            and (datetime.datetime.utcnow() - self.last_check).total_seconds() < one_day
        ):
            log.info("Update check performed in last day. Skipping.")
            return

        self.last_check = datetime.datetime.utcnow()
        if self.new_version is not None or self._check_updates():
            if not self.has_notified and settings.notify_updates:
                self.has_notified = True
                log.info("Update Available: {0}".format(self.new_version))
                self.playerManager.show_text(
                    _(
                        "MPV Shim v{0} Update Available\nOpen menu (press c) for details."
                    ).format(self.new_version),
                    5000,
                    1,
                )

    def open(self):
        self.playerManager.set_fullscreen(False)
        try:
            webbrowser.open(release_url + "latest")
        except Exception:
            log.error("Could not open release URL.", exc_info=True)



from PIL import Image
from collections import deque
import subprocess
from multiprocessing import Process, Queue
import threading
import sys
import logging
import queue

from .constants import USER_APP_NAME, APP_NAME
from .conffile import confdir
from .clients import clientManager
from .utils import get_resource
from .log_utils import CustomFormatter
from .i18n import _

log = logging.getLogger("gui_mgr")

# From https://stackoverflow.com/questions/6631299/
# This is for opening the config directory.


def _show_file_darwin(path: str):
    subprocess.Popen(["open", path])


def _show_file_linux(path: str):
    subprocess.Popen(["xdg-open", path])


def _show_file_win32(path: str):
    subprocess.Popen(["explorer", path])


_show_file_func = {
    "darwin": _show_file_darwin,
    "linux": _show_file_linux,
    "win32": _show_file_win32,
    "cygwin": _show_file_win32,
}

try:
    show_file = _show_file_func[sys.platform]

    def open_config():
        show_file(confdir(APP_NAME))


except KeyError:
    open_config = None
    log.warning("Platform does not support opening folders.")

# Setup a log handler for log items.
log_cache = deque([], 1000)
root_logger = logging.getLogger("")


class GUILogHandler(logging.Handler):
    def __init__(self):
        self.callback = None
        super().__init__()

    def emit(self, record):
        log_entry = self.format(record)
        log_cache.append(log_entry)

        if self.callback:
            try:
                self.callback(log_entry)
            except Exception:
                pass


guiHandler = GUILogHandler()
guiHandler.setFormatter(CustomFormatter())
root_logger.addHandler(guiHandler)

# Why am I using another process for the GUI windows?
# Because both pystray and tkinter must run
# in the main thread of their respective process.


class LoggerWindow(threading.Thread):
    def __init__(self):
        self.dead = False
        self.queue = None
        self.r_queue = None
        self.process = None
        threading.Thread.__init__(self)

    def run(self):
        self.queue = Queue()
        self.r_queue = Queue()
        self.process = LoggerWindowProcess(self.queue, self.r_queue)

        def handle(message):
            self.handle("append", message)

        self.process.start()
        handle("\n".join(log_cache))
        guiHandler.callback = handle
        while True:
            action, param = self.r_queue.get()
            if action == "die":
                self._die()
                break

    def handle(self, action: str, params=None):
        self.queue.put((action, params))

    def stop(self):
        self.r_queue.put(("die", None))

    def _die(self):
        guiHandler.callback = None
        self.handle("die")
        self.process.terminate()
        self.dead = True


class LoggerWindowProcess(Process):
    def __init__(self, queue: Queue, r_queue: Queue):
        self.queue = queue
        self.r_queue = r_queue
        self.tk = None
        self.root = None
        self.text = None
        Process.__init__(self)

    def update(self):
        try:
            self.text.config(state=self.tk.NORMAL)
            while True:
                action, param = self.queue.get_nowait()
                if action == "append":
                    self.text.config(state=self.tk.NORMAL)
                    self.text.insert(self.tk.END, "\n")
                    self.text.insert(self.tk.END, param)
                    self.text.config(state=self.tk.DISABLED)
                    self.text.see(self.tk.END)
                elif action == "die":
                    self.root.destroy()
                    self.root.quit()
                    return
        except queue.Empty:
            pass
        self.text.after(100, self.update)

    def run(self):
        import tkinter as tk

        self.tk = tk
        root = tk.Tk()
        self.root = root
        root.title(_("Application Log"))
        text = tk.Text(root)
        text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
        text.config(wrap=tk.WORD)
        self.text = text
        yscroll = tk.Scrollbar(command=text.yview)
        text["yscrollcommand"] = yscroll.set
        yscroll.pack(side=tk.RIGHT, fill=tk.Y)
        text.config(state=tk.DISABLED)
        self.update()
        root.mainloop()
        self.r_queue.put(("die", None))


class PreferencesWindow(threading.Thread):
    def __init__(self):
        self.dead = False
        self.dead_trigger = threading.Event()
        self.queue = None
        self.r_queue = None
        self.process = None
        threading.Thread.__init__(self)

    def run(self):
        self.queue = Queue()
        self.r_queue = Queue()
        self.process = PreferencesWindowProcess(self.queue, self.r_queue)
        self.process.start()
        self.handle("upd", clientManager.credentials)
        while True:
            action, param = self.r_queue.get()
            if action == "die":
                self._die()
                break
            elif action == "add":
                try:
                    is_logged_in = clientManager.login(*param)
                    if is_logged_in:
                        self.handle("upd", clientManager.credentials)
                    else:
                        self.handle("error")
                except Exception:
                    log.error("Error while adding server.", exc_info=True)
                    self.handle("error")
            elif action == "remove":
                clientManager.remove_client(param)
                self.handle("upd", clientManager.credentials)

    def handle(self, action: str, params=None):
        self.queue.put((action, params))

    def stop(self):
        self.r_queue.put(("die", None))

    def block_until_close(self):
        self.dead_trigger.wait()

    def _die(self):
        self.dead_trigger.set()
        self.handle("die")
        self.process.terminate()
        self.dead = True


class PreferencesWindowProcess(Process):
    def __init__(self, queue: Queue, r_queue: Queue):
        self.queue = queue
        self.r_queue = r_queue
        self.servers = None
        self.server_ids = None
        self.tk = None
        self.messagebox = None
        self.root = None
        self.serverList = None
        self.current_uuid = None
        self.servername = None
        self.username = None
        self.password = None
        self.add_button = None
        self.remove_button = None
        Process.__init__(self)

    def update(self):
        try:
            while True:
                action, param = self.queue.get_nowait()
                if action == "upd":
                    self.update_servers(param)
                    self.add_button.config(state=self.tk.NORMAL)
                    self.remove_button.config(state=self.tk.NORMAL)
                elif action == "error":
                    self.messagebox.showerror(
                        _("Add Server"),
                        _(
                            "Could not add server.\nPlease check your connection infomation."
                        ),
                    )
                    self.add_button.config(state=self.tk.NORMAL)
                elif action == "die":
                    self.root.destroy()
                    self.root.quit()
                    return
        except queue.Empty:
            pass
        self.root.after(100, self.update)

    def update_servers(self, server_list):
        self.servers = server_list
        self.server_ids = [x["uuid"] for x in self.servers]
        self.serverList.set(
            [
                "{0} ({1}, {2})".format(
                    server["Name"],
                    server["username"],
                    _("Ok") if server["connected"] else _("Fail"),
                )
                for server in self.servers
            ]
        )

    def run(self):
        import tkinter as tk
        from tkinter import ttk, messagebox

        self.tk = tk
        self.messagebox = messagebox
        root = tk.Tk()
        root.title(_("Server Configuration"))
        self.root = root

        self.servers = {}
        self.server_ids = []
        self.serverList = tk.StringVar(value=[])
        self.current_uuid = None

        def server_select(_x):
            idxs = serverlist.curselection()
            if len(idxs) == 1:
                self.current_uuid = self.server_ids[idxs[0]]

        c = ttk.Frame(root, padding=(5, 5, 12, 0))
        c.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        root.grid_columnconfigure(0, weight=1)
        root.grid_rowconfigure(0, weight=1)

        serverlist = tk.Listbox(c, listvariable=self.serverList, height=10, width=40)
        serverlist.grid(column=0, row=0, rowspan=6, sticky=(tk.N, tk.S, tk.E, tk.W))
        c.grid_columnconfigure(0, weight=1)
        c.grid_rowconfigure(4, weight=1)

        servername_label = ttk.Label(c, text=_("Server:"))
        servername_label.grid(column=1, row=0, sticky=tk.E)
        self.servername = tk.StringVar()
        servername_box = ttk.Entry(c, textvariable=self.servername)
        servername_box.grid(column=2, row=0)
        username_label = ttk.Label(c, text=_("Username:"))
        username_label.grid(column=1, row=1, sticky=tk.E)
        self.username = tk.StringVar()
        username_box = ttk.Entry(c, textvariable=self.username)
        username_box.grid(column=2, row=1)
        password_label = ttk.Label(c, text=_("Password:"))
        password_label.grid(column=1, row=2, sticky=tk.E)
        self.password = tk.StringVar()
        password_box = ttk.Entry(c, textvariable=self.password, show="*")
        password_box.grid(column=2, row=2)

        def add_server():
            self.add_button.config(state=tk.DISABLED)
            self.r_queue.put(
                (
                    "add",
                    (self.servername.get(), self.username.get(), self.password.get()),
                )
            )

        def remove_server():
            self.remove_button.config(state=tk.DISABLED)
            self.r_queue.put(("remove", self.current_uuid))

        def close():
            self.r_queue.put(("die", None))

        self.add_button = ttk.Button(c, text=_("Add Server"), command=add_server)
        self.add_button.grid(column=2, row=3, pady=5, sticky=tk.E)
        self.remove_button = ttk.Button(
            c, text=_("Remove Server"), command=remove_server
        )
        self.remove_button.grid(column=1, row=4, padx=5, pady=10, sticky=(tk.E, tk.S))
        close_button = ttk.Button(c, text=_("Close"), command=close)
        close_button.grid(column=2, row=4, pady=10, sticky=(tk.E, tk.S))

        serverlist.bind("<<ListboxSelect>>", server_select)
        self.update()
        root.mainloop()
        self.r_queue.put(("die", None))


# Q: OK. So you put Tkinter in it's own process.
#    Now why is Pystray in another process too?!
# A: Because if I don't, MPV and GNOME Appindicator
#    try to access the same resources and cause the
#    entire application to segfault.
#
# I suppose this means I can put the Tkinter GUI back
# into the main process. This is true, but then the
# two need to be merged, which is non-trivial.


class UserInterface(threading.Thread):
    def __init__(self):
        self.dead = False
        self.open_player_menu = lambda: None
        self.icon_stop = lambda: None
        self.log_window = None
        self.preferences_window = None
        self.stop_callback = None
        self.gui_ready = None
        self.r_queue = None
        self.process = None

        threading.Thread.__init__(self)

    def run(self):
        self.r_queue = Queue()
        self.process = STrayProcess(self.r_queue)
        self.process.start()

        while True:
            action, param = self.r_queue.get()
            if hasattr(self, action):
                getattr(self, action)()
            elif action == "die":
                self._die()
                if self.stop_callback:
                    self.stop_callback()
                break

    def stop(self):
        self.r_queue.put(("die", None))

    def _die(self):
        self.process.terminate()
        self.dead = True

        if self.log_window and not self.log_window.dead:
            self.log_window.stop()
        if self.preferences_window and not self.preferences_window.dead:
            self.preferences_window.stop()

    def login_servers(self):
        is_logged_in = clientManager.try_connect()
        if not is_logged_in:
            self.show_preferences()
            self.preferences_window.block_until_close()

    def show_console(self):
        if self.log_window is None or self.log_window.dead:
            self.log_window = LoggerWindow()
            self.log_window.start()

    def show_preferences(self):
        if self.preferences_window is None or self.preferences_window.dead:
            self.preferences_window = PreferencesWindow()
            self.preferences_window.start()

    def ready(self):
        if self.gui_ready:
            self.gui_ready.set()

    @staticmethod
    def open_config_brs():
        if open_config:
            open_config()
        else:
            log.error("Config opening is not available.")


class STrayProcess(Process):
    def __init__(self, r_queue: Queue):
        self.r_queue = r_queue
        self.icon_stop = None
        Process.__init__(self)

    def run(self):
        from pystray import Icon, MenuItem, Menu

        def get_wrapper(command):
            def wrapper():
                self.r_queue.put((command, None))

            return wrapper

        def die():
            # We don't call self.icon_stop() because it crashes on Linux now...
            if sys.platform == "linux":
                # This kills the status icon uncleanly.
                self.r_queue.put(("die", None))
            else:
                self.icon_stop()

        menu_items = [
            MenuItem(_("Configure Servers"), get_wrapper("show_preferences")),
            MenuItem(_("Show Console"), get_wrapper("show_console")),
            MenuItem(_("Application Menu"), get_wrapper("open_player_menu")),
            MenuItem(_("Open Config Folder"), get_wrapper("open_config_brs")),
            MenuItem(_("Quit"), die),
        ]

        icon = Icon(USER_APP_NAME, menu=Menu(*menu_items))
        icon.icon = Image.open(get_resource("systray.png"))
        self.icon_stop = icon.stop

        def setup(icon: Icon):
            icon.visible = True
            self.r_queue.put(("ready", None))

        icon.run(setup=setup)
        self.r_queue.put(("die", None))


user_interface = UserInterface()



import threading

from .player import playerManager


class ActionThread(threading.Thread):
    def __init__(self):
        self.trigger = threading.Event()
        self.halt = False

        threading.Thread.__init__(self)

    def stop(self):
        self.halt = True
        self.join()

    def run(self):
        force_next = False
        while not self.halt:
            if playerManager.is_active() or force_next:
                playerManager.update()

            force_next = False
            if self.trigger.wait(1):
                force_next = True
                self.trigger.clear()


actionThread = ActionThread()



from queue import Empty, Queue
import threading
import time
import urllib.request
from werkzeug.serving import make_server
from flask import Flask, request, jsonify
from time import sleep
import os.path
import json
import base64

# So, most of my use of the webview library is ugly but there's a reason for this!
# Debian's python3-webview package is super old (2.3), pip3's pywebview package is much newer (3.2).
# I would just say "fuck it, install from testing/backports/unstable" except that's not any newer either.
# So instead I'm going to try supporting *both* here, which gets rather ugly because they made very significant changes.
#
# The key differences seem to be:
# 3.2's create_window() returns a Window object immediately, then you need to call start()
# 2.3's create_window() blocks forever effectivelly calling start() itself
#
# 3.2's Window has a .loaded Event that you need to subscribe to notice when the window is ready for input
# 2.3 has a webview_ready() function that blocks until webview is ready (or timeout is passed)
import webview  # Python3-webview in Debian, pywebview in pypi
import sys
from threading import Event
import datetime

from ..clients import clientManager
from ..player import playerManager
from ..conf import settings
from ..event_handler import eventHandler
from ..constants import CAPABILITIES, CLIENT_VERSION, USER_APP_NAME, APP_NAME
from ..utils import get_resource
from .. import conffile
import logging

log = logging.getLogger("webclient")

remember_layout = conffile.get(APP_NAME, "layout.json")


def do_not_cache(response):
    response.cache_control.no_store = True
    if response.cache_control.max_age:
        response.cache_control.max_age = None
    if response.cache_control.public:
        response.cache_control.public = False


# Based on https://stackoverflow.com/questions/15562446/
class Server(threading.Thread):
    def __init__(self):
        self.srv = None
        self.ctx = None

        threading.Thread.__init__(self)

    def stop(self):
        if self.srv is not None:
            self.srv.shutdown()
        self.join()

    def run(self):
        app = Flask(
            __name__,
            static_url_path="",
            static_folder=get_resource("webclient_view", "webclient")
        )

        pl_event_queue = Queue()
        last_server_id = ""
        last_user_id = ""
        last_user_name = ""

        def wrap_playstate(active, playstate=None, item=None):
            if playstate is None:
                playstate = {
                    "CanSeek": False,
                    "IsPaused": False,
                    "IsMuted": False,
                    "RepeatMode": "RepeatNone"
                }
            res = {
                "PlayState": playstate,
                "AdditionalUsers": [],
                "Capabilities": {
                    "PlayableMediaTypes": CAPABILITIES["PlayableMediaTypes"].split(","),
                    "SupportedCommands": CAPABILITIES["SupportedCommands"].split(","),
                    "SupportsMediaControl": True,
                    "SupportsContentUploading": False,
                    "SupportsPersistentIdentifier": False,
                    "SupportsSync": False
                },
                "RemoteEndPoint": "0.0.0.0",
                "PlayableMediaTypes": CAPABILITIES["PlayableMediaTypes"].split(","),
                "Id": settings.client_uuid,
                "UserId": last_user_id,
                "UserName": last_user_name,
                "Client": USER_APP_NAME,
                "LastActivityDate": datetime.datetime.utcnow().isoformat(),
                "LastPlaybackCheckIn": "0001-01-01T00:00:00.0000000Z",
                "DeviceName": settings.player_name,
                "DeviceId": settings.client_uuid,
                "ApplicationVersion": CLIENT_VERSION,
                "IsActive": active,
                "SupportsMediaControl": True,
                "SupportsRemoteControl": True,
                "HasCustomDeviceName": False,
                "ServerId": last_server_id,
                "SupportedCommands": CAPABILITIES["SupportedCommands"].split(","),
                "dest": "player"
            }
            if "NowPlayingQueue" in playstate:
                res["NowPlayingQueue"] = playstate["NowPlayingQueue"]
            if "PlaylistItemId" in playstate:
                res["PlaylistItemId"] = playstate["PlaylistItemId"]
            if item:
                res["NowPlayingItem"] = item
            return res

        def on_playstate(state, payload=None, item=None):
            pl_event_queue.put(wrap_playstate(True, payload, item))
            if (state == "stopped"):
                pl_event_queue.put(wrap_playstate(False))

        def it_on_event(name, event):
            event = (event or {}).copy()
            event["Name"] = name
            event["dest"] = "ws"
            pl_event_queue.put(event)

        playerManager.on_playstate = on_playstate
        self.it_on_event = it_on_event
        self.it_event_set = {"UserDataChanged"}

        @app.after_request
        def add_header(response):
            if request.path == "/index.html":
                do_not_cache(response)
                client_data = base64.b64encode(json.dumps({
                    "appName": USER_APP_NAME,
                    "appVersion": CLIENT_VERSION,
                    "deviceName": settings.player_name,
                    "deviceId": settings.client_uuid
                }).encode('ascii'))
                # We need access to this data before we can make an async web call.
                replacement = b"""<body><script type="application/json" id="clientData">%s</script>""" % client_data
                if settings.desktop_scale != 1.0:
                    f_scale = float(settings.desktop_scale)
                    replacement = replacement + (b"""<style>body { zoom: %.2f; }</style>""" % f_scale)
                response.make_sequence()
                response.set_data(
                    response.get_data().replace(
                        b"<body>",
                        replacement,
                    )
                )

                return response
            if not response.cache_control.no_store:
                response.cache_control.max_age = 2592000
            return response

        @app.route("/mpv_shim_session", methods=["POST"])
        def mpv_shim_session():
            nonlocal last_server_id, last_user_id, last_user_name
            if request.headers["Content-Type"] != "application/json; charset=UTF-8":
                return "Go Away"
            req = request.json
            log.info("Recieved session for server: {0}, user: {1}".format(req["Name"], req["username"]))
            if req["Id"] not in clientManager.clients:
                is_logged_in = clientManager.connect_client(req)
                log.info("Connection was successful.")
            else:
                is_logged_in = True
                log.info("Ignoring as client already exists.")
            last_server_id = req["Id"]
            last_user_id = req["UserId"]
            last_user_name = req["username"]
            resp = jsonify({"success": is_logged_in})
            resp.status_code = 200 if is_logged_in else 500
            do_not_cache(resp)
            return resp

        @app.route("/mpv_shim_event", methods=["POST"])
        def mpv_shim_event():
            if request.headers["Content-Type"] != "application/json; charset=UTF-8":
                return "Go Away"
            try:
                queue_item = pl_event_queue.get(timeout=5)
            except Empty:
                queue_item = {}
            resp = jsonify(queue_item)
            resp.status_code = 200
            do_not_cache(resp)
            return resp

        @app.route("/mpv_shim_message", methods=["POST"])
        def mpv_shim_message():
            if request.headers["Content-Type"] != "application/json; charset=UTF-8":
                return "Go Away"
            req = request.json
            client = clientManager.clients.get(req["payload"]["ServerId"])
            resp = jsonify({})
            resp.status_code = 200
            do_not_cache(resp)
            if client is None:
                log.warning("Message recieved but no client available. Ignoring.")
                return resp
            # Assume only 1 client is connected.
            eventHandler.handle_event(client, req["name"], req["payload"])
            return resp

        @app.route("/mpv_shim_teardown", methods=["POST"])
        def mpv_shim_teardown():
            if request.headers["Content-Type"] != "application/json; charset=UTF-8":
                return "Go Away"
            log.info("Client teardown requested.")
            clientManager.stop_all_clients()
            resp = jsonify({})
            resp.status_code = 200
            do_not_cache(resp)
            return resp

        @app.route("/mpv_shim_syncplay_join", methods=["POST"])
        def mpv_shim_join():
            if request.headers["Content-Type"] != "application/json; charset=UTF-8":
                return "Go Away"
            req = request.json
            client = list(clientManager.clients.values())[0]
            playerManager.syncplay.client = client
            playerManager.syncplay.join_group(req["GroupId"])
            resp = jsonify({})
            resp.status_code = 200
            do_not_cache(resp)
            return resp

        self.srv = make_server("127.0.0.1", 18096, app, threaded=True)
        self.ctx = app.app_context()
        self.ctx.push()
        self.srv.serve_forever()


# This makes me rather uncomfortable, but there's no easy way around this other than
# importing display_mirror in helpers. Lambda needed because the 2.3 version of the JS
# api adds an argument even when not used.
class WebviewClient(object):
    def __init__(self, cef=False):
        self.open_player_menu = lambda: None
        self.server = Server()
        self.cef = cef
        self.webview = None

    def start(self):
        pass

    @staticmethod
    def login_servers():
        pass

    def get_webview(self):
        return self.webview

    def run(self):
        self.server.start()

        extra_options = {}
        if os.path.exists(remember_layout):
            with open(remember_layout) as fh:
                layout_options = json.load(fh)
            if (
                settings.desktop_keep_pos
                and layout_options.get("x")
                and layout_options.get("y")
            ):
                extra_options["x"] = layout_options["x"]
                extra_options["y"] = layout_options["y"]
            if (
                settings.desktop_keep_size
                and layout_options.get("width")
                and layout_options.get("height")
            ):
                extra_options["width"] = layout_options["width"]
                extra_options["height"] = layout_options["height"]
        else:
            # Set a reasonable window size
            extra_options.update({"width": 1280, "height": 720})
        if (
            not self.cef
            and sys.platform.startswith("win32")
            or sys.platform.startswith("cygwin")
        ):
            # I wasted half a day here. Turns out that pywebview does something that
            # breaks Jellyfin on Windows in both EdgeHTML and CEF. This kills that.
            try:
                from webview.platforms import winforms

                winforms.BrowserView.EdgeHTML.on_navigation_completed = lambda: None
            except Exception:
                pass

        # Apparently pywebview3 breaks everywhere, not just on Windows.
        if self.cef:
            try:
                from webview.platforms import cef

                cef.settings.update({"cache_path": conffile.get(APP_NAME, "cache")})
                cef.Browser.initialize = lambda self: None
            except Exception:
                pass
        try:
            from webview.platforms import cocoa

            def override_cocoa(_self, webview, _nav):
                # Add the webview to the window if it's not yet the contentView
                i = cocoa.BrowserView.get_instance("webkit", webview)

                if i:
                    if not webview.window():
                        i.window.setContentView_(webview)
                        i.window.makeFirstResponder_(webview)

            cocoa.BrowserView.BrowserDelegate.webView_didFinishNavigation_ = (
                override_cocoa
            )
        except Exception:
            pass
        try:
            from webview.platforms import gtk

            # noinspection PyUnresolvedReferences
            def override_gtk(_self, webview, _status):
                if not webview.props.opacity:
                    gtk.glib.idle_add(webview.set_opacity, 1.0)

            gtk.BrowserView.on_load_finish = override_gtk
        except Exception:
            pass
        try:
            from webview.platforms import qt

            qt.BrowserView.on_load_finished = lambda self: None
        except Exception:
            pass

        url = "http://127.0.0.1:18096/index.html"
        # Wait until the server is ready.
        while True:
            try:
                urllib.request.urlopen(url)
                break
            except Exception:
                pass
            sleep(0.1)

        # Webview needs to be run in the MainThread.
        window = webview.create_window(
            url=url,
            title="Jellyfin MPV Desktop",
            fullscreen=settings.desktop_fullscreen,
            **extra_options
        )
        if window is not None:
            self.webview = window

            def handle_close():
                x, y = window.x, window.y
                extra_options = {
                    "x": x,
                    "y": y,
                    "width": window.width,
                    "height": window.height,
                }
                with open(remember_layout, "w") as fh:
                    json.dump(extra_options, fh)

            window.closing += handle_close
            if self.cef:
                webview.start(gui="cef")
            else:
                webview.start()
        self.server.stop()

    def stop(self):
        self.server.stop()



from queue import LifoQueue
from .bulk_subtitle import process_series
from .conf import settings
from .utils import mpv_color_to_plex, get_sub_display_title
from .video_profile import VideoProfileManager
from .svp_integration import SVPManager
from .i18n import _

import time
import logging

log = logging.getLogger("menu")

TRANSCODE_LEVELS = (
    ("1080p 20 Mbps", 20000),
    ("1080p 12 Mbps", 12000),
    ("1080p 10 Mbps", 10000),
    ("720p 4 Mbps", 4000),
    ("720p 3 Mbps", 3000),
    ("720p 2.5 Mbps", 2500),
    ("540p 1.5 Mbps", 1500),
    ("540p 0.9 Mbps", 950),
    ("480p 0.4 Mbps", 400),
    ("320p 0.3 Mbps", 320),
)

COLOR_LIST = (
    (_("White"), "#FFFFFFFF"),
    (_("Yellow"), "#FFFFEE00"),
    (_("Black"), "#FF000000"),
    (_("Cyan"), "#FF00FFFF"),
    (_("Blue"), "#FF0000FF"),
    (_("Green"), "#FF00FF00"),
    (_("Magenta"), "#FFEE00EE"),
    (_("Red"), "#FFFF0000"),
    (_("Gray"), "#FF808080"),
)

SIZE_LIST = (
    (_("Tiny"), 50),
    (_("Small"), 75),
    (_("Normal"), 100),
    (_("Large"), 125),
    (_("Huge"), 200),
)

HEX_TO_COLOR = {v: c for c, v in COLOR_LIST}

lang_filter = set(settings.lang_filter.split(","))
if "und" in lang_filter:
    lang_filter.add(None)

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .player import playerManager as playerManager_type


class OSDMenu(object):
    def __init__(self, player_manager: "playerManager_type", player):
        self.playerManager = player_manager

        self.is_menu_shown = False
        self.menu_title = ""
        self.menu_stack = LifoQueue()
        self.menu_list = []
        self.menu_selection = 0
        self.menu_tmp = None
        self.mouse_back = False
        (
            self.original_osd_color,
            self.original_osd_size,
        ) = player_manager.get_osd_settings()

        self.profile_menu = None
        self.profile_manager = None
        if settings.shader_pack_enable:
            try:
                self.profile_manager = VideoProfileManager(self, player_manager, player)
                self.profile_menu = self.profile_manager.menu_action
            except Exception:
                log.error("Could not load profile manager.", exc_info=True)

        self.svp_menu = None
        try:
            self.svp_menu = SVPManager(self, player_manager)
        except Exception:
            log.error("Could not load SVP integration.", exc_info=True)

    # The menu is a bit of a hack...
    # It works using multiline OSD.
    # We also have to force the window to open.

    def refresh_menu(self):
        if not self.is_menu_shown:
            return

        items = self.menu_list
        selected_item = self.menu_selection

        if self.mouse_back:
            menu_text = "(<--) {0}".format(self.menu_title)
        else:
            menu_text = self.menu_title
        for i, item in enumerate(items):
            fmt = "\n   {0}"
            if i == selected_item and not self.mouse_back:
                fmt = "\n   **{0}**"
            menu_text += fmt.format(item[0])

        self.playerManager.show_text(menu_text, 2 ** 30, 1)

    def mouse_select(self, idx: int):
        if idx < 0 or idx > len(self.menu_list):
            return
        if idx == 0:
            self.mouse_back = True
        else:
            self.mouse_back = False
            self.menu_selection = idx - 1
        self.refresh_menu()

    def show_menu(self):
        self.is_menu_shown = True
        self.playerManager.set_osd_settings("#CC333333", 40)

        self.playerManager.enable_osc(False)
        self.playerManager.triggered_menu(True)

        self.menu_title = _("Main Menu")
        self.menu_selection = 0
        self.mouse_back = False

        if self.playerManager.is_playing():
            self.menu_list = [
                (_("Change Audio"), self.change_audio_menu),
                (_("Change Subtitles"), self.change_subtitle_menu),
                (_("Change Video Quality"), self.change_transcode_quality),
                (_("SyncPlay"), self.playerManager.syncplay.menu_action),
            ]
            if self.playerManager.update_check.new_version is not None:
                self.menu_list.insert(
                    0,
                    (
                        _("MPV Shim v{0} Release Info/Download").format(
                            self.playerManager.update_check.new_version
                        ),
                        self.playerManager.update_check.open,
                    ),
                )
            if self.profile_menu is not None:
                self.menu_list.append(
                    (_("Change Video Playback Profile"), self.profile_menu)
                )
            if self.playerManager.get_video().parent.is_tv:
                self.menu_list.append(
                    (
                        _("Auto Set Audio/Subtitles (Entire Series)"),
                        self.change_tracks_menu,
                    )
                )
            self.menu_list.append(
                (_("Quit and Mark Unwatched"), self.unwatched_menu_handle)
            )
            if settings.screenshot_menu:
                self.menu_list.append((_("Screenshot"), self.screenshot))
        else:
            self.menu_list = []
            if self.profile_menu is not None:
                self.menu_list.append((_("Video Playback Profiles"), self.profile_menu))

        if self.svp_menu is not None and self.svp_menu.is_available():
            self.menu_list.append((_("SVP Settings"), self.svp_menu.menu_action))

        self.menu_list.extend(
            [
                (_("Video Preferences"), self.video_preferences_menu),
                (_("Player Preferences"), self.player_preferences_menu),
                (_("Close Menu"), self.hide_menu),
            ]
        )

        self.refresh_menu()

        # Wait until the menu renders to pause.
        time.sleep(0.2)

        if self.playerManager.playback_is_aborted():
            self.playerManager.force_window(True)
        else:
            if not self.playerManager.syncplay.is_enabled():
                self.playerManager.set_paused(True)

    def hide_menu(self):
        if self.is_menu_shown:
            self.playerManager.set_osd_settings(
                self.original_osd_color, self.original_osd_size
            )
            self.playerManager.show_text("", 0, 0)

            self.playerManager.enable_osc(True)
            self.playerManager.triggered_menu(False)
            self.playerManager.force_window(False)

            if not self.playerManager.playback_is_aborted():
                if not self.playerManager.syncplay.is_enabled():
                    self.playerManager.set_paused(False)

        self.is_menu_shown = False

    def screenshot(self):
        self.playerManager.show_text("", 0, 0)
        time.sleep(0.5)
        self.playerManager.screenshot()
        self.hide_menu()

    def put_menu(self, title: str, entries: Optional[list] = None, selected: int = 0):
        if entries is None:
            entries = []

        self.menu_stack.put((self.menu_title, self.menu_list, self.menu_selection))
        self.menu_title = title
        self.menu_list = entries
        self.menu_selection = selected

    def menu_action(self, action: str):
        if not self.is_menu_shown and action in ("home", "ok"):
            self.show_menu()
        else:
            if action == "up":
                self.menu_selection = (self.menu_selection - 1) % len(self.menu_list)
            elif action == "down":
                self.menu_selection = (self.menu_selection + 1) % len(self.menu_list)
            elif action == "back":
                if self.menu_stack.empty():
                    self.hide_menu()
                else:
                    (
                        self.menu_title,
                        self.menu_list,
                        self.menu_selection,
                    ) = self.menu_stack.get_nowait()
            elif action == "ok":
                if self.mouse_back:
                    self.menu_action("back")
                else:
                    self.menu_list[self.menu_selection][1]()
            elif action == "home":
                self.show_menu()
            self.mouse_back = False
            self.refresh_menu()

    def change_audio_menu_handle(self):
        self.playerManager.put_task(
            self.playerManager.set_streams, self.menu_list[self.menu_selection][2], None
        )
        self.playerManager.timeline_handle()
        self.menu_action("back")

    def change_audio_menu(self):
        self.put_menu(_("Select Audio Track"))

        selected_aid = self.playerManager.get_video().aid
        audio_streams = [
            s
            for s in self.playerManager.get_video().media_source["MediaStreams"]
            if s.get("Type") == "Audio"
        ]
        for i, audio_track in enumerate(audio_streams):
            aid = audio_track.get("Index")
            if (
                settings.lang_filter_audio
                and aid != selected_aid
                and audio_track.get("Language") not in lang_filter
            ):
                continue

            self.menu_list.append(
                [
                    "{0} ({1})".format(
                        audio_track.get("DisplayTitle"), audio_track.get("Title")
                    ),
                    self.change_audio_menu_handle,
                    aid,
                ]
            )
            if aid == selected_aid:
                self.menu_selection = i

    def change_subtitle_menu_handle(self):
        self.playerManager.put_task(
            self.playerManager.set_streams, None, self.menu_list[self.menu_selection][2]
        )
        self.playerManager.timeline_handle()
        self.menu_action("back")

    def change_subtitle_menu(self):
        self.put_menu(_("Select Subtitle Track"))

        selected_sid = self.playerManager.get_video().sid
        subtitle_streams = [
            s
            for s in self.playerManager.get_video().media_source["MediaStreams"]
            if s.get("Type") == "Subtitle"
        ]
        self.menu_list.append([_("None"), self.change_subtitle_menu_handle, -1])
        for i, subtitle_track in enumerate(subtitle_streams):
            sid = subtitle_track.get("Index")
            if (
                settings.lang_filter_sub
                and sid != selected_sid
                and subtitle_track.get("Language") not in lang_filter
            ):
                continue

            self.menu_list.append(
                [
                    "{0} ({1})".format(
                        get_sub_display_title(subtitle_track),
                        subtitle_track.get("Title"),
                    ),
                    self.change_subtitle_menu_handle,
                    sid,
                ]
            )
            if sid == selected_sid:
                self.menu_selection = i + 1

    def change_transcode_quality_handle(self):
        bitrate = self.menu_list[self.menu_selection][2]
        if bitrate == "none":
            self.playerManager.get_video().set_trs_override(None, False)
        elif bitrate == "max":
            self.playerManager.get_video().set_trs_override(None, True)
        else:
            self.playerManager.get_video().set_trs_override(bitrate, True)

        self.menu_action("back")
        self.playerManager.put_task(self.playerManager.restart_playback)
        self.playerManager.timeline_handle()

    def change_transcode_quality(self):
        handle = self.change_transcode_quality_handle
        self.put_menu(
            _("Select Transcode Quality"),
            [(_("No Transcode"), handle, "none"), (_("Maximum"), handle, "max")],
        )

        for item in TRANSCODE_LEVELS:
            self.menu_list.append((item[0], handle, item[1]))

        self.menu_selection = 7
        cur_bitrate = self.playerManager.get_video().get_transcode_bitrate()
        for i, option in enumerate(self.menu_list):
            if cur_bitrate == option[2]:
                self.menu_selection = i

    def change_tracks_handle(self):
        mode = self.menu_list[self.menu_selection][2]
        process_series(mode, self.playerManager)

    def change_tracks_manual_s1(self):
        self.change_audio_menu()
        for item in self.menu_list:
            item[1] = self.change_tracks_manual_s2

    def change_tracks_manual_s2(self):
        self.menu_tmp = self.menu_selection
        self.change_subtitle_menu()
        for item in self.menu_list:
            item[1] = self.change_tracks_manual_s3

    def change_tracks_manual_s3(self):
        aid, sid = self.menu_tmp, self.menu_selection - 1
        # Pop 3 menu items.
        for i in range(3):
            self.menu_action("back")
        process_series("manual", self.playerManager, aid, sid)

    def change_tracks_menu(self):
        self.put_menu(
            _("Select Audio/Subtitle for Series"),
            [
                (_("English Audio"), self.change_tracks_handle, "dubbed"),
                (
                    _("Japanese Audio w/ English Subtitles"),
                    self.change_tracks_handle,
                    "subbed",
                ),
                (
                    _("Manual by Track Index (Less Reliable)"),
                    self.change_tracks_manual_s1,
                ),
            ],
        )

    def settings_toggle_bool(self):
        _x, _x, key, name = self.menu_list[self.menu_selection]
        setattr(settings, key, not getattr(settings, key))
        settings.save()
        self.menu_list[self.menu_selection] = self.get_settings_toggle(name, key)

    def get_settings_toggle(self, name: str, setting: str):
        return (
            "{0}: {1}".format(name, getattr(settings, setting)),
            self.settings_toggle_bool,
            setting,
            name,
        )

    def transcode_settings_handle(self):
        settings.remote_kbps = self.menu_list[self.menu_selection][2]
        settings.save()

        # Need to re-render preferences menu.
        for i in range(2):
            self.menu_action("back")
        self.video_preferences_menu()

    def transcode_settings_menu(self):
        self.put_menu(_("Select Default Transcode Profile"))
        handle = self.transcode_settings_handle
        self.menu_list.append((_("No Transcode") + " 2 Gbps", handle, 2147483))

        for i, item in enumerate(TRANSCODE_LEVELS):
            self.menu_list.append((item[0], handle, item[1]))
            if settings.remote_kbps == item[1]:
                self.menu_selection = i

    @staticmethod
    def get_subtitle_color(color: str):
        if color in HEX_TO_COLOR:
            return HEX_TO_COLOR[color]
        else:
            return mpv_color_to_plex(color)

    def sub_settings_handle(self):
        setting_name = self.menu_list[self.menu_selection][2]
        value = self.menu_list[self.menu_selection][3]
        setattr(settings, setting_name, value)
        settings.save()

        # Need to re-render preferences menu.
        for i in range(2):
            self.menu_action("back")
        self.video_preferences_menu()

        if self.playerManager.get_video().is_transcode:
            if setting_name == "subtitle_size":
                self.playerManager.put_task(self.playerManager.update_subtitle_visuals)
        else:
            self.playerManager.update_subtitle_visuals()

    def subtitle_color_menu(self):
        self.put_menu(
            _("Select Subtitle Color"),
            [
                (name, self.sub_settings_handle, "subtitle_color", color)
                for name, color in COLOR_LIST
            ],
        )

    def subtitle_size_menu(self):
        self.put_menu(
            _("Select Subtitle Size"),
            [
                (name, self.sub_settings_handle, "subtitle_size", size)
                for name, size in SIZE_LIST
            ],
            selected=2,
        )

    def subtitle_position_menu(self):
        self.put_menu(
            _("Select Subtitle Position"),
            [
                (_("Bottom"), self.sub_settings_handle, "subtitle_position", "bottom"),
                (_("Top"), self.sub_settings_handle, "subtitle_position", "top"),
                (_("Middle"), self.sub_settings_handle, "subtitle_position", "middle"),
            ],
        )

    def video_preferences_menu(self):
        self.put_menu(
            _("Video Preferences"),
            [
                (
                    _("Remote Transcode Quality: {0:0.1f} Mbps").format(
                        settings.remote_kbps / 1000
                    ),
                    self.transcode_settings_menu,
                ),
                (
                    _("Subtitle Size: {0}").format(settings.subtitle_size),
                    self.subtitle_size_menu,
                ),
                (
                    _("Subtitle Position: {0}").format(settings.subtitle_position),
                    self.subtitle_position_menu,
                ),
                (
                    _("Subtitle Color: {0}").format(
                        self.get_subtitle_color(settings.subtitle_color)
                    ),
                    self.subtitle_color_menu,
                ),
                self.get_settings_toggle(_("Transcode H265 to H264"), "transcode_h265"),
                self.get_settings_toggle(
                    _("Transcode Hi10p to 8bit"), "transcode_hi10p"
                ),
                self.get_settings_toggle(_("Direct Paths"), "direct_paths"),
                self.get_settings_toggle(_("Transcode to H265"), "transcode_to_h265"),
                self.get_settings_toggle(_("Disable Direct Play"), "always_transcode"),
            ],
        )

    def player_preferences_menu(self):
        self.put_menu(
            _("Player Preferences"),
            [
                self.get_settings_toggle(_("Auto Play"), "auto_play"),
                self.get_settings_toggle(_("Auto Fullscreen"), "fullscreen"),
                self.get_settings_toggle(_("Media Key Seek"), "media_key_seek"),
                self.get_settings_toggle(_("Use Web Seek Pref"), "use_web_seek"),
                self.get_settings_toggle(_("Display Mirroring"), "display_mirroring"),
                self.get_settings_toggle(_("Write Logs to File"), "write_logs"),
                self.get_settings_toggle(_("Check for Updates"), "check_updates"),
                self.get_settings_toggle(
                    _("Discord Rich Presence"), "discord_presence"
                ),
            ],
        )

    def unwatched_menu_handle(self):
        self.playerManager.put_task(self.playerManager.unwatched_quit)
        self.playerManager.timeline_handle()



from jellyfin_apiclient_python import JellyfinClient
from jellyfin_apiclient_python.connection_manager import CONNECTION_STATE
from .conf import settings
from . import conffile
from getpass import getpass
from .constants import CAPABILITIES, CLIENT_VERSION, USER_APP_NAME, USER_AGENT, APP_NAME
from .i18n import _

import sys
import os.path
import json
import uuid
import time
import logging
import re

log = logging.getLogger("clients")
path_regex = re.compile("^(https?://)?([^/:]+)(:[0-9]+)?(/.*)?$")

from typing import Optional


def expo(max_value: Optional[int] = None):
    n = 0
    while True:
        a = 2 ** n
        if max_value is None or a < max_value:
            yield a
            n += 1
        else:
            yield max_value


class ClientManager(object):
    def __init__(self):
        self.callback = lambda client, event_name, data: None
        self.credentials = []
        self.clients = {}
        self.usernames = {}
        self.is_stopping = False

    def cli_connect(self):
        is_logged_in = self.try_connect()
        add_another = False

        if "add" in sys.argv:
            add_another = True

        while not is_logged_in or add_another:
            server = input(_("Server URL: "))
            username = input(_("Username: "))
            password = getpass(_("Password: "))

            is_logged_in = self.login(server, username, password)

            if is_logged_in:
                log.info(_("Successfully added server."))
                add_another = input(_("Add another server?") + " [y/N] ")
                add_another = add_another in ("y", "Y", "yes", "Yes")
            else:
                log.warning(_("Adding server failed."))

    @staticmethod
    def client_factory():
        client = JellyfinClient(allow_multiple_clients=True)
        client.config.data["app.default"] = True
        client.config.app(
            USER_APP_NAME, CLIENT_VERSION, settings.player_name, settings.client_uuid
        )
        client.config.data["http.user_agent"] = USER_AGENT
        client.config.data["auth.ssl"] = not settings.ignore_ssl_cert
        return client

    def _connect_all(self):
        is_logged_in = False
        for server in self.credentials:
            if self.connect_client(server):
                is_logged_in = True
        return is_logged_in

    def try_connect(self):
        credentials_location = conffile.get(APP_NAME, "cred.json")
        if os.path.exists(credentials_location):
            with open(credentials_location) as cf:
                self.credentials = json.load(cf)

        if "Servers" in self.credentials:
            credentials_old = self.credentials
            self.credentials = []
            for server in credentials_old["Servers"]:
                server["uuid"] = str(uuid.uuid4())
                server["username"] = ""
                self.credentials.append(server)

        is_logged_in = self._connect_all()
        if settings.connect_retry_mins and not is_logged_in:
            log.warning(
                "Connection failed. Will retry for {0} minutes.".format(
                    settings.connect_retry_mins
                )
            )
            for attempt in range(settings.connect_retry_mins * 2):
                time.sleep(30)
                is_logged_in = self._connect_all()
                if is_logged_in:
                    break

        return is_logged_in

    def save_credentials(self):
        credentials_location = conffile.get(APP_NAME, "cred.json")
        with open(credentials_location, "w") as cf:
            json.dump(self.credentials, cf)

    def login(
        self, server: str, username: str, password: str, force_unique: bool = False
    ):
        protocol, host, port, path = path_regex.match(server).groups()

        if not protocol:
            log.warning("Adding http:// because it was not provided.")
            protocol = "http://"

        if protocol == "http://" and not port:
            log.warning("Adding port 8096 for insecure local http connection.")
            log.warning(
                "If you want to connect to standard http port 80, use :80 in the url."
            )
            port = ":8096"

        server = "".join(filter(bool, (protocol, host, port, path)))

        client = self.client_factory()
        client.auth.connect_to_address(server)
        result = client.auth.login(server, username, password)
        if "AccessToken" in result:
            credentials = client.auth.credentials.get_credentials()
            server = credentials["Servers"][0]
            if force_unique:
                server["uuid"] = server["Id"]
            else:
                server["uuid"] = str(uuid.uuid4())
            server["username"] = username
            if force_unique and server["Id"] in self.clients:
                return True
            self.connect_client(server)
            self.credentials.append(server)
            self.save_credentials()
            return True
        return False

    def setup_client(self, client: "JellyfinClient", server):
        def event(event_name, data):
            if event_name == "WebSocketDisconnect":
                timeout_gen = expo(100)
                if server["uuid"] in self.clients:
                    while not self.is_stopping:
                        timeout = next(timeout_gen)
                        log.info(
                            "No connection to server. Next try in {0} second(s)".format(
                                timeout
                            )
                        )
                        self._disconnect_client(server=server)
                        time.sleep(timeout)
                        if self.connect_client(server):
                            break
            else:
                self.callback(client, event_name, data)

        client.callback = event
        client.callback_ws = event
        client.start(websocket=True)

        client.jellyfin.post_capabilities(CAPABILITIES)

    def remove_client(self, uuid: str):
        self.credentials = [
            server for server in self.credentials if server["uuid"] != uuid
        ]
        self.save_credentials()
        self._disconnect_client(uuid=uuid)

    def connect_client(self, server):
        if self.is_stopping:
            return False

        is_logged_in = False
        client = self.client_factory()
        state = client.authenticate({"Servers": [server]}, discover=False)
        server["connected"] = state["State"] == CONNECTION_STATE["SignedIn"]
        if server["connected"]:
            is_logged_in = True
            self.clients[server["uuid"]] = client
            self.setup_client(client, server)
            if server.get("username"):
                self.usernames[server["uuid"]] = server["username"]

        return is_logged_in

    def _disconnect_client(self, uuid: Optional[str] = None, server=None):
        if uuid is None and server is not None:
            uuid = server["uuid"]

        if uuid not in self.clients:
            return

        if server is not None:
            server["connected"] = False

        client = self.clients[uuid]
        del self.clients[uuid]
        client.stop()

    def remove_all_clients(self):
        self.stop_all_clients()
        self.credentials = []
        self.save_credentials()

    def stop_all_clients(self):
        for key, client in list(self.clients.items()):
            del self.clients[key]
            client.stop()

    def stop(self):
        self.is_stopping = True
        for client in self.clients.values():
            client.stop()

    def get_username_from_client(self, client):
        # This is kind of convoluted. It may fail if a server
        # was added before we started saving usernames.
        for uuid, client2 in self.clients.items():
            if client2 is client:
                if uuid in self.usernames:
                    return self.usernames[uuid]
                for server in self.credentials:
                    if server["uuid"] == uuid:
                        return server.get("username", "Unknown")
                break

        return "Unknown"


clientManager = ClientManager()



import logging
import re

bad_patterns = (
    (re.compile("api_key=[a-f0-9]*"), "api_key=REDACTED"),
    (
        re.compile("'X-MediaBrowser-Token': '[a-f0-9]*'"),
        "'X-MediaBrowser-Token': 'REDACTED'",
    ),
    (re.compile("'AccessToken': '[a-f0-9]*'"), "'AccessToken': 'REDACTED'"),
)

sanitize_logs = False
root_logger = logging.getLogger("")
root_logger.level = logging.DEBUG


def sanitize(message):
    if type(message) in (int, float):
        return message
    if message is not str and message is not bytes:
        message = str(message)
    for pattern, replacement in bad_patterns:
        message = pattern.sub(replacement, message)
    return message


class CustomFormatter(logging.Formatter):
    def __init__(self, force_sanitize=False):
        self.force_sanitize = force_sanitize
        super(CustomFormatter, self).__init__(
            fmt="%(asctime)s [%(levelname)8s] %(name)s: %(message)s"
        )

    def format(self, record):
        if sanitize_logs or self.force_sanitize:
            record.msg = sanitize(record.msg)
            if type(record.args) is dict:
                sanitized = {}
                for key, value in record.args.items():
                    sanitized[key] = sanitize(value)
                record.args = sanitized
            elif type(record.args) is tuple:
                record.args = tuple(sanitize(value) for value in record.args)
            else:
                return "(Log message could not be processed for sanitization.)"
        return logging.Formatter.format(self, record)


def enable_sanitization():
    global sanitize_logs
    sanitize_logs = True


def configure_log(destination):
    handler = logging.StreamHandler(destination)
    handler.setFormatter(CustomFormatter())
    root_logger.addHandler(handler)


def configure_log_file(destination: str):
    handler = logging.FileHandler(destination, mode="w")
    # Never allow logging API keys to a file.
    handler.setFormatter(CustomFormatter(True))
    root_logger.addHandler(handler)



from .clients import clientManager


class UserInterface(object):
    def __init__(self):
        self.open_player_menu = lambda: None
        self.stop = lambda: None

    @staticmethod
    def login_servers():
        clientManager.cli_connect()

    def start(self):
        pass

    def stop(self):
        pass


user_interface = UserInterface()



# noinspection PyUnresolvedReferences,PyPackageRequirements
import win32gui
import logging

log = logging.getLogger("win_utils")


def window_enumeration_handler(hwnd, top_windows):
    top_windows.append((hwnd, win32gui.GetWindowText(hwnd)))


def raise_mpv():
    # This workaround is madness. Apparently SetForegroundWindow
    # won't work randomly, so I have to call ShowWindow twice.
    # Once to hide the window, and again to successfully raise the window.
    try:
        top_windows = []
        fg_win = win32gui.GetForegroundWindow()
        win32gui.EnumWindows(window_enumeration_handler, top_windows)
        for i in top_windows:
            if " - mpv" in i[1].lower():
                if i[0] != fg_win:
                    win32gui.ShowWindow(i[0], 6)  # Minimize
                    win32gui.ShowWindow(i[0], 9)  # Un-minimize
                break

    except Exception:
        log.error("Could not raise MPV.", exc_info=True)


def mirror_act(state: bool, name: str = "Jellyfin MPV Shim Mirror"):
    try:
        top_windows = []
        win32gui.EnumWindows(window_enumeration_handler, top_windows)
        for i in top_windows:
            if name in i[1]:
                print(i)
                win32gui.ShowWindow(i[0], 9 if state else 6)
                break

    except Exception:
        log.error("Could not raise/lower MPV mirror.", exc_info=True)



from collections import namedtuple
from .utils import get_sub_display_title
from .i18n import _

# TODO: Should probably support automatic profiles for languages other than English and Japanese...

import time
import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .player import PlayerManager as PlayerManager_type

Part = namedtuple("Part", ["id", "audio", "subtitle"])
Audio = namedtuple("Audio", ["id", "language_code", "name", "display_name"])
Subtitle = namedtuple(
    "Subtitle", ["id", "language_code", "name", "is_forced", "display_name"]
)

log = logging.getLogger("bulk_subtitle")
messages = []
keep_messages = 6


def render_message(message, show_text):
    log.info(message)
    messages.append(message)
    text = _("Selecting Tracks...")
    for message in messages[-6:]:
        text += "\n   " + message
    show_text(text, 2 ** 30, 1)


def process_series(mode, player: "PlayerManager_type", m_raid=None, m_rsid=None):
    messages.clear()
    media = player.get_video()
    client = media.client
    show_text = player.show_text
    c_aid, c_sid = None, None
    c_pid = media.media_source.get("Id")

    success_ct = 0
    partial_ct = 0
    count = 0

    videos = client.jellyfin.get_season(media.item["SeriesId"], media.item["SeasonId"])[
        "Items"
    ]
    for video in videos:
        name = "s{0}e{1:02}".format(
            video.get("ParentIndexNumber"), video.get("IndexNumber")
        )
        video = client.jellyfin.get_item(video.get("Id"))
        for media_source in video["MediaSources"]:
            count += 1
            audio_list = [
                Audio(
                    s.get("Index"),
                    s.get("Language"),
                    s.get("Title"),
                    s.get("DisplayTitle"),
                )
                for s in media_source["MediaStreams"]
                if s.get("Type") == "Audio"
            ]
            subtitle_list = [
                Subtitle(
                    s.get("Index"),
                    s.get("Language"),
                    s.get("Title"),
                    s.get("IsForced"),
                    get_sub_display_title(s),
                )
                for s in media_source["MediaStreams"]
                if s.get("Type") == "Subtitle"
            ]
            part = Part(media_source.get("Id"), audio_list, subtitle_list)

            aid = None
            sid = -1
            if mode == "subbed":
                audio, subtitle = get_subbed(part)
                if audio and subtitle:
                    render_message(
                        "{0}: {1} ({2})".format(
                            name, subtitle.display_name, subtitle.name
                        ),
                        show_text,
                    )
                    aid, sid = audio.id, subtitle.id
                    success_ct += 1
            elif mode == "dubbed":
                audio, subtitle = get_dubbed(part)
                if audio and subtitle:
                    render_message(
                        "{0}: {1} ({2})".format(
                            name, subtitle.display_name, subtitle.name
                        ),
                        show_text,
                    )
                    aid, sid = audio.id, subtitle.id
                    success_ct += 1
                elif audio:
                    render_message(_("{0}: No Subtitles").format(name), show_text)
                    aid = audio.id
                    partial_ct += 1
            elif mode == "manual":
                if m_raid < len(part.audio) and m_rsid < len(part.subtitle):
                    audio = part.audio[m_raid]
                    aid = audio.id
                    render_message(
                        "{0} a: {1} ({2})".format(name, audio.display_name, audio.name),
                        show_text,
                    )
                    if m_rsid != -1:
                        subtitle = part.subtitle[m_rsid]
                        sid = subtitle.id
                        render_message(
                            "{0} s: {1} ({2})".format(
                                name, subtitle.display_name, subtitle.name
                            ),
                            show_text,
                        )
                    success_ct += 1

            if aid:
                if c_pid == part.id:
                    c_aid, c_sid = aid, sid

                # This is a horrible hack to change the audio/subtitle settings without playing
                # the media. I checked the jelyfin source code and didn't find a better way.
                client.jellyfin.session_progress(
                    {
                        "ItemId": part.id,
                        "AudioStreamIndex": aid,
                        "SubtitleStreamIndex": sid,
                    }
                )

            else:
                render_message(_("{0}: Fail").format(name), show_text)

    if mode == "subbed":
        render_message(
            _("Set Subbed: {0} ok, {1} fail").format(success_ct, count - success_ct),
            show_text,
        )
    elif mode == "dubbed":
        render_message(
            _("Set Dubbed: {0} ok, {1} audio only, {2} fail").format(
                success_ct, partial_ct, count - success_ct - partial_ct
            ),
            show_text,
        )
    elif mode == "manual":
        render_message(
            _("Manual: {0} ok, {1} fail").format(success_ct, count - success_ct),
            show_text,
        )
    time.sleep(3)
    if c_aid:
        render_message(_("Setting Current..."), show_text)
        player.put_task(player.set_streams, c_aid, c_sid)
        player.timeline_handle()


def get_subbed(part):
    japanese_audio = None
    english_subtitles = None
    subtitle_weight = None

    for audio in part.audio:
        lower_title = audio.name.lower() if audio.name is not None else ""
        if audio.language_code != "jpn" and "japan" not in lower_title:
            continue
        if "commentary" in lower_title:
            continue

        if japanese_audio is None:
            japanese_audio = audio
            break

    for subtitle in part.subtitle:
        lower_title = subtitle.name.lower() if subtitle.name is not None else ""
        if subtitle.language_code != "eng" and "english" not in lower_title:
            continue
        if subtitle.is_forced:
            continue

        weight = dialogue_weight(lower_title)
        if subtitle_weight is None or weight < subtitle_weight:
            subtitle_weight = weight
            english_subtitles = subtitle

    if japanese_audio and english_subtitles:
        return japanese_audio, english_subtitles
    return None, None


def get_dubbed(part):
    english_audio = None
    sign_subtitles = None
    subtitle_weight = None

    for audio in part.audio:
        lower_title = audio.name.lower() if audio.name is not None else ""
        if audio.language_code != "eng" and "english" not in lower_title:
            continue
        if "commentary" in lower_title:
            continue

        if english_audio is None:
            english_audio = audio
            break

    for subtitle in part.subtitle:
        lower_title = subtitle.name.lower() if subtitle.name is not None else ""
        if subtitle.language_code != "eng" and "english" not in lower_title:
            continue
        if subtitle.is_forced:
            sign_subtitles = subtitle
            break

        weight = sign_weight(lower_title)
        if weight == 0:
            continue

        if subtitle_weight is None or weight < subtitle_weight:
            subtitle_weight = weight
            sign_subtitles = subtitle

    if english_audio:
        return english_audio, sign_subtitles
    return None, None


def dialogue_weight(text: str):
    if not text:
        return 900
    lower_text = text.lower()
    has_dialogue = (
        "main" in lower_text or "full" in lower_text or "dialogue" in lower_text
    )
    has_songs = "op/ed" in lower_text or "song" in lower_text or "lyric" in lower_text
    has_signs = "sign" in lower_text
    vendor = "bd" in lower_text or "retail" in lower_text
    weight = 900

    if has_dialogue and has_songs:
        weight -= 100
    if has_songs:
        weight += 200
    if has_dialogue and has_signs:
        weight -= 100
    elif has_signs:
        weight += 700
    if vendor:
        weight += 50
    return weight


def sign_weight(text: str):
    if not text:
        return 0
    lower_text = text.lower()
    has_songs = "op/ed" in lower_text or "song" in lower_text or "lyric" in lower_text
    has_signs = "sign" in lower_text
    vendor = "bd" in lower_text or "retail" in lower_text
    weight = 900

    if not (has_songs or has_signs):
        return 0
    if has_songs:
        weight -= 200
    if has_signs:
        weight -= 300
    if vendor:
        weight += 50
    return weight



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="jellyfin-mpv-shim",
    version="1.10.0",
    author="Ian Walton",
    author_email="iwalton3@gmail.com",
    description="Cast media from Jellyfin Mobile and Web apps to MPV.",
    license="GPLv3",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jellyfin/jellyfin-desktop",
    packages=[
        "jellyfin_mpv_shim",
        "jellyfin_mpv_shim.display_mirror",
        "jellyfin_mpv_shim.webclient_view",
    ],
    package_data={
        "jellyfin_mpv_shim.display_mirror": ["*.css", "*.html"],
        "jellyfin_mpv_shim": ["systray.png"],
    },
    entry_points={
        "console_scripts": [
            "jellyfin-mpv-shim=jellyfin_mpv_shim.mpv_shim:main",
            "jellyfin-mpv-desktop=jellyfin_mpv_shim.mpv_shim:main_desktop",
            "jellyfin-desktop=jellyfin_mpv_shim.mpv_shim:main_desktop",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    extras_require={
        "gui": ["pystray", "PIL"],
        "mirror": ["Jinja2", "pywebview>=3.3.1"],
        "desktop": ["Flask", "pywebview>=3.3.1", "Werkzeug"],
        "discord": ["pypresence"],
        "all": [
            "Jinja2",
            "pywebview>=3.3.1",
            "pystray",
            "Flask",
            "Werkzeug",
            "pypresence",
        ],
    },
    python_requires=">=3.6",
    install_requires=[
        "python-mpv",
        "jellyfin-apiclient-python>=1.7.0",
        "python-mpv-jsonipc>=1.1.9",
        "requests",
        "pydantic",
    ],
    include_package_data=True,
)



import os.path
import os
import sys
import getpass

# If no platform is matched, use the current directory.
_confdir = None
username = getpass.getuser()


def posix(app: str):
    if os.environ.get("XDG_CONFIG_HOME"):
        return os.path.join(os.environ["XDG_CONFIG_HOME"], app)
    else:
        return os.path.join(os.path.expanduser("~"), ".config", app)


def win32(app: str):
    if os.environ.get("APPDATA"):
        return os.path.join(os.environ["APPDATA"], app)
    else:
        return os.path.join(r"C:\Users", username, r"AppData\Roaming", app)


confdirs = (
    ("linux", posix),
    ("win32", win32),
    ("cygwin", posix),
    (
        "darwin",
        lambda app: os.path.join(
            "/Users", username, "Library/Application Support", app
        ),
    ),
)

for platform, directory in confdirs:
    if sys.platform.startswith(platform):
        _confdir = directory

custom_config = None
for i, arg in enumerate(sys.argv):
    if arg == "--config" and len(sys.argv) > i + 1:
        custom_config = sys.argv[i + 1]


def confdir(app: str):
    if custom_config is not None:
        return custom_config
    elif _confdir is not None:
        return _confdir(app)
    else:
        return ""


def get(app: str, conf_file: str, create: bool = False):
    conf_folder = confdir(app)
    if not os.path.isdir(conf_folder):
        os.makedirs(conf_folder)
    conf_file = os.path.join(conf_folder, conf_file)
    if create and not os.path.isfile(conf_file):
        open(conf_file, "w").close()
    return conf_file
